/**
 * @file usb.c
 *
 * @brief This file contains the usb stack for the USBlini firmware
 *        project.
 *
 * @author Thomas Fischl (https://www.fischl.de/contact.html)
 * @copyright (c) 2021 Thomas Fischl
 * 
 * This file is part of the USBlini firmware project and is copyrighted by the
 * authors listed above. It is free for private or educational non-commercial use
 * (for a commercial license please contact the authors). It may not be
 * redistributed without the prior written consent of the authors.
 *
 * It is provided "as is" without warranty of any kind, either express or implied,
 * including without limitation any implied warranties of condition, uninterrupted
 * use, merchantability, fitness for a particular purpose, or non-infringement. In
 * no event shall the authors be liable for any direct or indirect damages arising
 * in any way out of the use of it.
 *
 */

#include "usblini.h"
#include "usb_descr.h"
#include "statusreport.h"
#include "usb.h"

volatile BDT epbd[EPBD_NROF] __at(0x2000);

volatile unsigned char ep0out_buffer[EP_BUFFERSIZE] __at(0x2028);
volatile unsigned char ep0in_buffer[EP_BUFFERSIZE] __at(0x2030);
volatile unsigned char ep1in_buffer[2][EP_BULK_BUFFERSIZE] __at(0x2038);
volatile unsigned char ep2in_buffer[2][EP_BULK_BUFFERSIZE] __at(0x20b8);

unsigned short usb_sendleft = 0;
unsigned char usb_sendpending = 0;
unsigned char usb_usertransfer = 0;
unsigned short usb_usertransfer_length = 0;

const unsigned char * usb_sendbuffer;
unsigned configured = 0;
unsigned char usb_config = 0;
unsigned char usb_setaddress = 0;
unsigned char usb_ep0status[2];

#define EVEN 0
#define ODD 1
unsigned char current_ep1_buffer = EVEN;
unsigned char current_ep2_buffer = EVEN;
unsigned char ep1_position = 0;
unsigned char ep2_position = 0;


/**
 * @brief Initialize USB stack
 */
void usb_init(void) {
    epbd[EPBD_EP0_OUT].stat = 0x80;
    epbd[EPBD_EP0_OUT].cnt = EP_BUFFERSIZE;
    epbd[EPBD_EP0_OUT].adrl = 0x28;
    epbd[EPBD_EP0_OUT].adrh = 0x20;

    epbd[EPBD_EP0_IN].stat = 0;
    epbd[EPBD_EP0_IN].cnt = EP_BUFFERSIZE;
    epbd[EPBD_EP0_IN].adrl = 0x30;
    epbd[EPBD_EP0_IN].adrh = 0x20;

    epbd[EPBD_EP1_IN_EVEN].stat = 0x40;
    epbd[EPBD_EP1_IN_EVEN].cnt = EP_BULK_BUFFERSIZE;
    epbd[EPBD_EP1_IN_EVEN].adrl = 0x38;
    epbd[EPBD_EP1_IN_EVEN].adrh = 0x20;

    epbd[EPBD_EP1_IN_ODD].stat = 0x40;
    epbd[EPBD_EP1_IN_ODD].cnt = EP_BULK_BUFFERSIZE;
    epbd[EPBD_EP1_IN_ODD].adrl = 0x78;
    epbd[EPBD_EP1_IN_ODD].adrh = 0x20;

    epbd[EPBD_EP2_IN_EVEN].stat = 0x40;
    epbd[EPBD_EP2_IN_EVEN].cnt = EP_BULK_BUFFERSIZE;
    epbd[EPBD_EP2_IN_EVEN].adrl = 0xb8;
    epbd[EPBD_EP2_IN_EVEN].adrh = 0x20;

    epbd[EPBD_EP2_IN_ODD].stat = 0x40;
    epbd[EPBD_EP2_IN_ODD].cnt = EP_BULK_BUFFERSIZE;
    epbd[EPBD_EP2_IN_ODD].adrl = 0xf8;
    epbd[EPBD_EP2_IN_ODD].adrh = 0x20;
  
    
    UEP0 = 0x16; // enable IN, enable OUT, enable CONTROL, enable handshake
    UEP1 = 0x1A; // enable IN, disable OUT, disable CONTROL, enable handshake
    UEP2 = 0x1A; // enable IN, disable OUT, disable CONTROL, enable handshake

    UCFG = 0x17; // enable pullup and full-speed, ping-pong on ep1-ep15
    UCON = 0x08; // enable usb module

    usb_ep0status[0] = 0;
    usb_ep0status[1] = 0;
}

/**
 * @brief Shutdown USB stack
 */
void usb_shutdown(void) {
    UCON = 0x00;
    UCFG = 0x00;
}

/**
 * @brief Determine if serial number is available
 * @return 0 if no number available
 */
unsigned char usb_serialNumberAvailable(void) {
    return ((usb_string_serial[0] == USB_STRING_SERIALNUMBER_SIZE) && (usb_string_serial[1] == 0x03));
}

/**
 * @brief Process pending send activity
 */
void usb_sendProcess(void) {
    if (!usb_sendpending) return;

    unsigned short length = usb_sendleft;
    if (length < EP_BUFFERSIZE) {
        // last packet
        usb_sendpending = 0; 
    } else {
        length = EP_BUFFERSIZE;
    }

    unsigned char i;
    for (i = 0; i < length; i++) {

        if ((usb_sendbuffer == usb_dev_desc + USB_DEV_DESC_SERIALNUMBER_OFFSET) && usb_serialNumberAvailable()) ep0in_buffer[i] = USB_STRING_SERIALNUMBER_INDEX;
        else ep0in_buffer[i] = *usb_sendbuffer;

        usb_sendbuffer++;
        usb_sendleft--;
    }

    epbd[EPBD_EP0_IN].cnt = (unsigned char)length;
    if (epbd[EPBD_EP0_IN].stat & 0x40)
        epbd[EPBD_EP0_IN].stat = 0x88;
    else
        epbd[EPBD_EP0_IN].stat = 0xC8;
}

/**
 * @brief Load given descriptor buffer into send buffer
 *
 * @param descbuffer Pointer to descriptor
 * @param size Size of descriptor
 * @param length Count of bytes to send
 * @return 0 on error, 1 on success
 */
unsigned char usb_loadDescriptor(const unsigned char * descbuffer, unsigned short size, unsigned short length) {

    if (length > size)
        length = size;

    usb_sendleft = length;
    usb_sendbuffer = descbuffer;
    usb_sendpending = 1;

    usb_sendProcess();
    return (unsigned char) length;
}

/**
 * @brief Handle descriptor requests
 *
 * @param type Type of descriptor
 * @param index Index of descriptor
 * @param length Requested length
 * @return 0 on error, 1 on success
 */
unsigned char usb_handleDescriptorRequest(unsigned char type, unsigned char index, unsigned short length) {

    switch (type) {
        case DESCR_DEVICE:
            return usb_loadDescriptor(usb_dev_desc, sizeof (usb_dev_desc), length);
        case DESCR_CONFIG:
            return usb_loadDescriptor(usb_config_desc, sizeof (usb_config_desc), length);
        case DESCR_STRING:
            switch (index) {
                case 0: return usb_loadDescriptor(usb_string_language, sizeof (usb_string_language), length);
                case 1: return usb_loadDescriptor(usb_string_manuf, sizeof (usb_string_manuf), length);
                case 2: return usb_loadDescriptor(usb_string_product, sizeof (usb_string_product), length);
                case 3: return usb_loadDescriptor(usb_string_serial, USB_STRING_SERIALNUMBER_SIZE, length);
            }
    }

    return 0;
}

/**
 * @brief Do IN transfer
 */
void ut_in(void) {

    unsigned short length = usb_usertransfer_length;
    if (length > EP_BUFFERSIZE)
        length = EP_BUFFERSIZE;

    length = usb_vendor_in((unsigned char*) ep0in_buffer, (unsigned char)length);

    usb_usertransfer_length -= length;

    epbd[EPBD_EP0_IN].cnt = (unsigned char)length;
    if (epbd[EPBD_EP0_IN].stat & 0x40)
        epbd[EPBD_EP0_IN].stat = 0x88;
    else
        epbd[EPBD_EP0_IN].stat = 0xC8;

    if (length < EP_BUFFERSIZE) {
        usb_usertransfer = 0;
    }
}

/**
 * @brief Trigger send out of EP1 buffer
 * @param length Length of buffer to send
 */
void usb_ep1_arm(unsigned char length) {

    if (current_ep1_buffer == EVEN) {
        epbd[EPBD_EP1_IN_EVEN].stat = 0x88;
        epbd[EPBD_EP1_IN_EVEN].cnt = length;
        current_ep1_buffer = ODD;
    } else {
        epbd[EPBD_EP1_IN_ODD].stat = 0xC8;
        epbd[EPBD_EP1_IN_ODD].cnt = length;
        current_ep1_buffer = EVEN;
    }        
    
}

/**
 * @brief Get pointer to current EP1 buffer
 * @return EP1 buffer
 */
volatile unsigned char * usb_ep1_getBuffer(void) {
    return ep1in_buffer[current_ep1_buffer];
}

/**
 * @brief Put given data into EP1 buffer
 * @param data Data for EP1 buffer
 * @param len Data length
 */
void usb_ep1_put(unsigned char * data, unsigned char len) {
    
    if (!configured) return;
    
    if (epbd[EPBD_EP1_IN_EVEN + current_ep1_buffer].stat & 0x80) {
        statusreport_setErrorflag(ERRORFLAG_EP1_OVERFLOW);
        return;
    }
    
    unsigned char i;
    for (i = 0; i < len; i++) {
        ep1in_buffer[current_ep1_buffer][ep1_position++] = data[i];
    }
    
    if (ep1_position >= EP_BULK_BUFFERSIZE) {
        ep1_position = 0;
        usb_ep1_arm(EP_BULK_BUFFERSIZE);
    }
}

/**
 * @brief Send out EP1 if there are data in buffer and the other ping pong
 * buffer is available.
 */
void usb_ep1_transmit(void) {
    if (ep1_position == 0) return;
    if (epbd[EPBD_EP1_IN_EVEN].stat & 0x80) return;
    if (epbd[EPBD_EP1_IN_ODD].stat & 0x80) return;
    
    usb_ep1_arm(ep1_position);
    ep1_position = 0;
}

/**
 * @brief Trigger send out of EP2 buffer
 */
void usb_ep2_arm(void) {
    
    if (current_ep2_buffer == EVEN) {
        epbd[EPBD_EP2_IN_EVEN].stat = 0x88;
        current_ep2_buffer = ODD;
    } else {
        epbd[EPBD_EP2_IN_ODD].stat = 0xC8;
        current_ep2_buffer = EVEN;
    }        
    
}

/**
 * @brief Put given character into EP2 buffer
 * @param ch Character to send via EP2 buffer
 */
void usb_ep2_put(unsigned char ch) {
    
    if (!configured) return;
    if (epbd[EPBD_EP2_IN_EVEN + current_ep2_buffer].stat & 0x80) return;
        
    ep2in_buffer[current_ep2_buffer][ep2_position] = ch;
    ep2_position++;
    if (ep2_position >= EP_BULK_BUFFERSIZE) {
        ep2_position = 0;
        usb_ep2_arm();
    }
}


/**
 * @brief Do USB stack processing
 */
void usb_process(void) {

    usb_ep1_transmit();
    
    if (UIRbits.TRNIF) {
        // complete interrupt

        if (USTAT == USTAT_EP0_OUT) {

            // out/setup

            if (((epbd[EPBD_EP0_OUT].stat >> 2) & 0x0F) == USB_PID_SETUP) {
                // setup token

                usb_usertransfer = 0;

                // stop pending/old/prepared IN transactions
                epbd[EPBD_EP0_IN].stat = 0;
                usb_sendpending = 0;

                if ((ep0out_buffer[0] & USBRQ_TYPE_MASK) == USBRQ_TYPE_STANDARD) {

                    switch (ep0out_buffer[1]) {
                        case REQUEST_GET_DESCRIPTOR:
                            if (!usb_handleDescriptorRequest(ep0out_buffer[3], ep0out_buffer[2], (unsigned short)((ep0out_buffer[7] << 8) | ep0out_buffer[6]))) {
                                epbd[EPBD_EP0_IN].cnt = 0;
                                epbd[EPBD_EP0_IN].stat = 0xCC; // Stall
                            }
                            break;
                        case REQUEST_SET_ADDRESS:

                            usb_setaddress = ep0out_buffer[2];

                            epbd[EPBD_EP0_IN].cnt = 0;
                            epbd[EPBD_EP0_IN].stat = 0xC8;

                            break;

                        case REQUEST_SET_CONFIGURATION:

                            usb_config = ep0out_buffer[2];
                            configured = 1;
                            epbd[EPBD_EP0_IN].cnt = 0;
                            epbd[EPBD_EP0_IN].stat = 0xC8;
                            break;

                        case REQUEST_GET_CONFIGURATION:

                            ep0in_buffer[0] = usb_config;
                            epbd[EPBD_EP0_IN].cnt = 1;
                            epbd[EPBD_EP0_IN].stat = 0xC8;
                            break;

                        case REQUEST_GET_INTERFACE:

                            ep0in_buffer[0] = 0;
                            epbd[EPBD_EP0_IN].cnt = 1;
                            epbd[EPBD_EP0_IN].stat = 0xC8;
                            break;

                        case REQUEST_GET_STATUS:
                            if ((ep0out_buffer[0] & USBRQ_RECIPIENT_MASK) == USBRQ_RECIPIENT_ENDPOINT) {
                                ep0in_buffer[0] = usb_ep0status[0];
                                ep0in_buffer[1] = usb_ep0status[1];
                            } else {
                                ep0in_buffer[0] = 0;
                                ep0in_buffer[1] = 0;
                            }
                            epbd[EPBD_EP0_IN].cnt = 2;
                            epbd[EPBD_EP0_IN].stat = 0xC8;
                            break;

                        case REQUEST_SET_FEATURE:
                            if ((ep0out_buffer[0] & USBRQ_RECIPIENT_MASK) == USBRQ_RECIPIENT_ENDPOINT) {
                                if (ep0out_buffer[2] == 0x00) // HALT
                                    usb_ep0status[0] = 1;
                            }
                            epbd[EPBD_EP0_IN].cnt = 0;
                            epbd[EPBD_EP0_IN].stat = 0xC8;
                            break;

                        case REQUEST_CLEAR_FEATURE:
                            if ((ep0out_buffer[0] & USBRQ_RECIPIENT_MASK) == USBRQ_RECIPIENT_ENDPOINT) {
                                usb_ep0status[0] = 0;                                                                
                            }
                            epbd[EPBD_EP0_IN].cnt = 0;
                            epbd[EPBD_EP0_IN].stat = 0xC8;
                            break;

                        case REQUEST_SYNCH_FRAME:
                            ep0in_buffer[0] = 0;
                            ep0in_buffer[1] = 0;
                            epbd[EPBD_EP0_IN].cnt = 2;
                            epbd[EPBD_EP0_IN].stat = 0xC8;
                            break;

                        case REQUEST_SET_INTERFACE:
                            epbd[EPBD_EP0_IN].cnt = 0;
                            epbd[EPBD_EP0_IN].stat = 0xC8;
                            break;

                        default:
                            epbd[EPBD_EP0_IN].cnt = 0;
                            epbd[EPBD_EP0_IN].stat = 0xCC; // stall
                            break;

                    }
                } else {

                    // not standard type -> class or vendor

                    if ((ep0out_buffer[0] & 0x80) == 0x80) {
                        // vendor or class control in

                        unsigned char len = usb_vendor_setup((unsigned char *) ep0out_buffer, (unsigned char *) ep0in_buffer);

                        if (len == 0xff) {
                            usb_usertransfer_length = (unsigned short)((ep0out_buffer[7] << 8) | ep0out_buffer[6]);
                            usb_usertransfer = 1;
                            ut_in();
                        } else {
                            epbd[EPBD_EP0_IN].cnt = len;
                            epbd[EPBD_EP0_IN].stat = 0xC8;
                        }


                    } else {
                        // vendor or class control out

                        unsigned char len = usb_vendor_setup((unsigned char *) ep0out_buffer, (unsigned char *) ep0in_buffer);
                        if (len == 0xff) {
                            usb_usertransfer_length = (unsigned short)((ep0out_buffer[7] << 8) | ep0out_buffer[6]);
                            usb_usertransfer = 2;
                        } else {
                            // TODO: what to do here? is zero package correct?
                            epbd[EPBD_EP0_IN].cnt = 0;
                            epbd[EPBD_EP0_IN].stat = 0xC8;
                        }
                    }
                }

            } else {
                // USB_PID_OUT

                // stop pending/old/prepared IN transactions
                epbd[EPBD_EP0_IN].stat = 0;
                usb_sendpending = 0;

                if (usb_usertransfer == 2) {

                    unsigned char ack = usb_vendor_out((unsigned char*) ep0out_buffer, epbd[EPBD_EP0_OUT].cnt);

                    if (ack) {
                        epbd[EPBD_EP0_IN].cnt = 0;
                        epbd[EPBD_EP0_IN].stat = 0xC8;
                    } else {
                        epbd[EPBD_EP0_IN].cnt = 0;
                        epbd[EPBD_EP0_IN].stat = 0xCC; // Stall
                    }
                }

            }

            epbd[EPBD_EP0_OUT].cnt = EP_BUFFERSIZE;
            epbd[EPBD_EP0_OUT].stat = 0x80;


        } else if (USTAT == USTAT_EP0_IN) {

            // check if set address command is pending
            if (usb_setaddress > 0) {
                UADDR = usb_setaddress;
                usb_setaddress = 0;
            }

            if (usb_usertransfer == 1) {
                ut_in();
            } else {
                usb_sendProcess();
            }
        }

        UCONbits.PKTDIS = 0;
        UIRbits.TRNIF = 0;
    }
}
